package org.livee.data.query;
/*
 * Copyright (c) 2007 Hideya Ochiai,  All rights reserved.
 * 
 * Permission of redistribution and use in source and binary forms, 
 * with or without modification, are granted, free of charge, to any person 
 * obtaining the copy of this software under the following conditions:
 * 
 *  1. Any copies of this source code must include the above copyright notice,
 *  this permission notice and the following statement without modification 
 *  except possible additions of other copyright notices. 
 * 
 *  2. Redistributions of the binary code must involve the copy of the above 
 *  copyright notice, this permission notice and the following statement 
 *  in documents and/or materials provided with the distribution.  
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

import java.sql.Connection;
import java.util.Map;

import org.livee.data.db.IDbCommand;
import org.livee.data.expr.AbstractData;
import org.json.JSONObject;
import org.json.JSONArray;

import org.livee.data.tools.NeighborManager;

/**
 * 
 * @author Hideya Ochiai, The University of Tokyo
 */
public class QueryProcessCommand implements IDbCommand {

	AbstractData m_Query;
	AbstractData m_QueryResponse;
	boolean m_bSucceeded;
	
	public QueryProcessCommand(AbstractData query){
		m_Query=query;
		m_bSucceeded=false;
	}
	
	public boolean succeeded(){
		return m_bSucceeded;
	}
	
	public AbstractData getQueryResponse(){
		return m_QueryResponse;
	}
	
	public void execCommand(Connection connection, Map conf) throws Exception {
		
		DataFactory df=DataFactory.getInstance();
		m_QueryResponse=df.create("queryResponse");
		
		java.sql.Statement stmt=null;
		try{
			stmt=connection.createStatement();
			if(m_Query.getDataType().equals("Query")){
				
				boolean includeChildren=false;
				String targetSite=null;
				String namePredicate=m_Query.getAttribute("name");
				
				if(namePredicate.startsWith("*.")){
					includeChildren=true;
					targetSite=namePredicate.substring(2);
				}else if(namePredicate.equals("*")){
					includeChildren=true;
					targetSite="";
				}else{
					targetSite=namePredicate;
				}
				
				// myself & parent resolution;
				String mySite=null;
				String mySiteServer=null;
				String parentSite=null;
				String parentSiteServer=null;
				String sql="SELECT name,type,server FROM admin.neighbor WHERE type=\'colleague\' OR type=\'parent\';";
				java.sql.ResultSet rset=stmt.executeQuery(sql);
				while(rset.next()){
					String name=rset.getString(1);
					String type=rset.getString(2);
					String server=rset.getString(3);
					if(type.equals("colleague")){
						mySite=name;
						mySiteServer=server;
					}else if(type.equals("parent")){
						parentSite=name;
						parentSiteServer=server;
					}
				}
				rset.close();
				
				boolean includeParent=false;
				boolean bMatches=false;
				if(includeChildren){
					if(targetSite.equals(mySite)){
						bMatches=true;
					}else if(mySite.endsWith(targetSite)){
						bMatches=true;
						includeParent=true;
					}
				}else if(targetSite.equals(mySite)){
					bMatches=true;
				}
			
				// If predicateName matches this server
				//   ==> Add myself to response
				//   ==> parentMatch too? ==> Add parent to redirect
				//   ==> childrenMatch too? ==> Add children that satisfy the filters to redirect
				// Else If
				//   ==> Add parent to redirect OR Drill Down to a child
				// End If
				
				// If predicateName matches this server
				if(bMatches){	
					AbstractData adminSite=null;
					boolean hasRedirect=false;
					
					AbstractData redirect=df.create("redirectTo");
					if(includeParent && parentSite!=null){ // If parent matches too ==> Add parent to redirect
						adminSite=createAdminSite(df,parentSite,null,parentSiteServer);
						redirect.append(adminSite);
						hasRedirect=true;
					}
					
					if(includeChildren){
						sql=((QueryData)m_Query).generateSQL();
						rset=stmt.executeQuery(sql);
						while(rset.next()){
							String name=rset.getString("name");
							String server=rset.getString("server");
							if(!name.equals(mySite)){
								adminSite=createAdminSite(df,name,null,server);
								redirect.append(adminSite);
								hasRedirect=true;
							}else{
								// ==> Add myself to response;
								adminSite=createAdminSite(df,mySite,null,mySiteServer);
								m_QueryResponse.append(adminSite);
							}
						}
					}
					
					if(!includeParent && !includeChildren){ // ExactMatch
						// ==> Add myself to response;
						adminSite=createAdminSite(df,mySite,null,mySiteServer);
						m_QueryResponse.append(adminSite);
					}
					
					if(hasRedirect){
						m_QueryResponse.append(redirect);
					}
				}else{ // Else (matches...) ==> Add parent to redirect
					boolean hasRedirect=false;
					AbstractData redirect=df.create("redirectTo");
					if(targetSite.endsWith(mySite)){
						// Go Down to a target child
						String nextTargetSite=null;
						int prefixIndex=targetSite.lastIndexOf(mySite);
						if(prefixIndex>=2){
							int nextPrefixIndexA=targetSite.lastIndexOf('.',prefixIndex-2);
							if(nextPrefixIndexA>=0){
								nextTargetSite=targetSite.substring(nextPrefixIndexA+1);
							}else{
								nextTargetSite=targetSite;
							}	
						}
						
						if(nextTargetSite!=null){
							sql="SELECT name,server FROM admin.neighbor WHERE name=\'"+nextTargetSite+"\';";
							rset=stmt.executeQuery(sql);
							if(rset.next()){
								String name=rset.getString(1);
								String server=rset.getString(2);
								AbstractData adminSite=createAdminSite(df,name,null,server);
								redirect.append(adminSite);
								hasRedirect=true;
							}
							rset.close();
						}
						
					}else{
						// Go up
						if(parentSite!=null){
							AbstractData adminSite=createAdminSite(df,parentSite,null,parentSiteServer);
							redirect.append(adminSite);
							hasRedirect=true;
						}
					}
					
					if(hasRedirect){
						m_QueryResponse.append(redirect);
					}
				}
			}
			m_bSucceeded=true;
		}
		catch(Exception e){
			e.printStackTrace();
		}
		finally{
			stmt.close();
		}
		
	}
	public String getCommandName() {
		return "QueryProcess";
	}
	public short getImplVersion() {
		return 0x0100;
	}
	
	private AbstractData createAdminSite(DataFactory df,String name,String type,String server){
		AbstractData adminSite=df.create("adminSite");
		adminSite.setAttribute("name", name);
		adminSite.setAttribute("type", type);
		try{
			JSONArray server_array=new JSONArray(server);
			int len=server_array.length();
			for(int i=0;i<len;i++){
				AbstractData adminService=df.create("adminServer");
				JSONObject obj=server_array.getJSONObject(i);
				
				String service=obj.getString("service");
				String url=obj.getString("url");
				String linkStatus=null;
				if(!obj.isNull("linkStatus")){
					linkStatus=obj.getString("linkStatus");
				}
				adminService.setAttribute("service", service);
				adminService.setAttribute("url", url);
				adminService.setAttribute("linkStatus", linkStatus);
				adminSite.append(adminService);
			}
		}
		catch(Exception e){
			e.printStackTrace();
		}
		return adminSite;
	}
}